from typing import Text, List, Tuple, Union, Dict, Generator

import pytest

from alembic.operations import Operations
from alembic.runtime.migration import MigrationContext
import sqlalchemy as sa
from sqlalchemy.engine import RowProxy
from sqlalchemy.orm import Session

from rasax.community.database.schema_migrations.alembic import utils as migration_utils


@pytest.fixture
def alembic_op(session: Session) -> Generator[Operations, None, None]:
    migration_context = MigrationContext.configure(session.get_bind())

    op = Operations(migration_context)

    # Replace `op` import for the time of the tests.
    old_op = migration_utils.op
    migration_utils.op = op

    yield op

    # Revert replacing of the `op` import
    migration_utils.op = old_op


def _insert_test_data(
    table: sa.Table, session: Session, rows: List[Union[Tuple, Dict]]
) -> None:
    """Insert values in existing table."""

    for values in rows:
        session.execute(table.insert().values(values))
    session.commit()


def test_modify_columns(alembic_op: Operations, session: Session):
    table_name = "test_table"
    test_bool_column = "test_bool_column"
    test_string_column = "test_string_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_bool_column, sa.Boolean, nullable=False),
        sa.Column(test_string_column, sa.String, nullable=False),
    )

    table = migration_utils.get_reflected_table(table_name, alembic_op)

    # Insert fake data
    _insert_test_data(
        table, session, [(1, False, "some text"), (2, True, "some other text")]
    )

    # Assert fake data is there
    entries = session.execute(table.select()).fetchall()
    assert entries == [(1, False, "some text"), (2, True, "some other text")]

    # Transform columns and values in column
    test_bool_transformation = migration_utils.ColumnTransformation(
        test_bool_column,
        new_column_kwargs={"type_": sa.String()},
        modify_from_column_value=lambda x: str(x) + "_test",
    )
    test_string_transformation = migration_utils.ColumnTransformation(
        test_string_column,
        new_column_args=[sa.Text()],
        new_column_kwargs={"nullable": True},
    )

    migration_utils.modify_columns(
        table_name, [test_bool_transformation, test_string_transformation]
    )

    # Get updated table reflection
    table = migration_utils.get_reflected_table(table_name, alembic_op)

    # Test if new value is expected
    entries = session.execute(table.select()).fetchall()

    assert entries == [
        (1, "False_test", "some text"),
        (2, "True_test", "some other text"),
    ]


def test_modify_columns_from_row_with_copy(alembic_op: Operations, session: Session):
    table_name = "test_modify_columns_from_row_with_copy"
    test_bool_column = "test_bool_column"
    test_string_column = "test_string_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_bool_column, sa.Boolean, nullable=False),
        sa.Column(test_string_column, sa.String, nullable=False),
    )

    # Get reflected table
    table = migration_utils.get_reflected_table(table_name, alembic_op)

    # Insert fake data
    _insert_test_data(
        table, session, [(1, False, "some text"), (2, True, "some other text")]
    )

    def modify_from_row(row: RowProxy) -> Text:
        return f"bool: {row.test_bool_column}, text: {row.test_string_column}"

    # Transform columns and values in column
    new_column_name = "text_bool_combined"
    test_bool_transformation = migration_utils.ColumnTransformation(
        test_bool_column,
        new_column_kwargs={"type_": sa.String()},
        modify_from_row=modify_from_row,
        new_column_name=new_column_name,
    )
    migration_utils.modify_columns(table_name, [test_bool_transformation])

    # Update table reflection
    table = migration_utils.get_reflected_table(table_name, alembic_op)

    # Test if new value is expected
    entries = session.execute(table.select()).fetchall()

    assert entries == [
        (1, False, "some text", "bool: False, text: some text"),
        (2, True, "some other text", "bool: True, text: some other text"),
    ]


def test_existing_table_exists(alembic_op: Operations, session: Session):
    table_name = "existing_table"
    test_column = "test_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_column, sa.String, nullable=False),
    )

    assert migration_utils.table_exists(table_name)


def test_nonexistent_table_doesnt_exist(alembic_op: Operations, session: Session):
    assert not migration_utils.table_exists("nonexistent_table")


def test_table_has_existing_column(alembic_op: Operations, session: Session):
    table_name = "test_column_table"
    test_column = "existing_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_column, sa.String, nullable=False),
    )

    assert migration_utils.table_has_column(table_name, test_column)


def test_table_doesnt_have_nonexistent_column(alembic_op: Operations, session: Session):
    table_name = "test_column_table_2"
    test_column = "existing_column"

    # Create test table
    alembic_op.create_table(
        table_name,
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(test_column, sa.String, nullable=False),
    )

    assert not migration_utils.table_has_column(table_name, "nonexistent_column")


def test_set_synonym_references_for_old_mapped_values(
    alembic_op: Operations, session: Session
):
    metadata = sa.MetaData(bind=alembic_op.get_bind())
    metadata.reflect()

    # Insert entity
    entity_table = migration_utils.get_reflected_table(
        "nlu_training_data_entity", session
    )
    _insert_test_data(entity_table, session, [dict(original_value="hut")])

    # Insert synonym
    synonym_table = migration_utils.get_reflected_table("entity_synonym", session)
    _insert_test_data(
        synonym_table, session, [dict(id=1, name="house", project_id="default")]
    )

    # Insert mapping
    synonym_value_table = migration_utils.get_reflected_table(
        "entity_synonym_value", session
    )
    _insert_test_data(
        synonym_value_table, session, [dict(entity_synonym_id=1, name="hut")]
    )

    # Run migration function
    from rasax.community.database.schema_migrations.alembic.versions import (
        migration_2019_10_25_add_column_for_synonym_foreign_key_a813c6c15a4c as migration,
    )

    migration.set_synonym_references_for_old_mapped_values(session)

    # Check if value was updated
    rows = session.execute(entity_table.select()).fetchall()

    assert rows[0]["entity_synonym_id"] == 1


def test_add_unique_id_to_entitysynonyms_migration(session_without_migration: Session):
    from rasax.community import sql_migrations

    alembic_config = sql_migrations._get_alembic_config(session_without_migration)
    import rasax.community.database.schema_migrations.alembic.versions.migration_2019_08_26_oracle_migrations_4daabca814ee as previous_migration

    # Run alembic migrations until the one to test
    sql_migrations._run_alembic_migration(alembic_config, previous_migration.revision)

    # Insert fake project to satisfy the foreign key constraint
    project_table = migration_utils.get_reflected_table(
        "project", session_without_migration
    )
    project_name = "new project"
    _insert_test_data(
        project_table,
        session_without_migration,
        [dict(project_id=project_name, team="A-team")],
    )

    # Insert a synonym with two values
    synonym_table = migration_utils.get_reflected_table(
        "entity_synonym", session_without_migration
    )
    name_of_synonym = "a great synonym"
    _insert_test_data(
        synonym_table,
        session_without_migration,
        [
            dict(synonym=name_of_synonym, name="test", project_id=project_name),
            dict(synonym=name_of_synonym, name="demo", project_id=project_name),
        ],
    )

    # Run migration to test
    import rasax.community.database.schema_migrations.alembic.versions.migration_2019_08_30_add_unique_id_to_entitysynonyms_d45a0bf21e89 as new_migration

    sql_migrations._run_alembic_migration(alembic_config, new_migration.revision)

    # Check if migration was successful
    synonym_table = migration_utils.get_reflected_table(
        "entity_synonym", session_without_migration
    )
    synonyms = session_without_migration.execute(synonym_table.select()).fetchall()

    # Check if synonyms were migrated correctly
    assert len(synonyms) == 1
    assert synonyms[0].name == name_of_synonym
    assert synonyms[0].project_id == project_name

    synonym_value_table = migration_utils.get_reflected_table(
        "entity_synonym_value", session_without_migration
    )
    synonym_values = session_without_migration.execute(
        synonym_value_table.select()
    ).fetchall()

    # Check if synonym values were migrated correctly
    assert len(synonym_values) == 2
    assert {value.name for value in synonym_values} == {"test", "demo"}
    assert all([value.entity_synonym_id == synonyms[0].id for value in synonym_values])
